package version10_2_2_1.Interface;

import java.util.ArrayList;

/**
 * This class define the initial set of primitive interactions
 * @author simon gay
 */
public class InteractionList {
	
	private ArrayList<PrimitiveAction> actions;
	private ArrayList<PrimitivePerception> perception;
	
	private ArrayList<PrimitiveInteraction> primitiveList;
	
	public static int size=25;					// width of visual system
	
	public static int nbDF=size*size *6 *4;		// number of secondary (visual) interactions
	public static int length=nbDF +7;				// 6 "primary" interactions >O, >P, >X, >F, v, ^ + 1 bias
	
	public InteractionList(){
		
		// define here the possible actions
		actions=new ArrayList<PrimitiveAction>();
		actions.add(new PrimitiveAction(">",0));
		actions.add(new PrimitiveAction("v",1));
		actions.add(new PrimitiveAction("^",2));
		
		// define the perception vector(s)
		perception=new ArrayList<PrimitivePerception>();
		perception.add(new PrimitivePerception("O",0));
		perception.add(new PrimitivePerception("P",1));
		perception.add(new PrimitivePerception("X",2));
		perception.add(new PrimitivePerception("F",3));
		
		// define the set of primitive interactions
		primitiveList=new ArrayList<PrimitiveInteraction>();
		primitiveList.add(new PrimitiveInteraction(actions.get(0),perception.get(0)));
		primitiveList.add(new PrimitiveInteraction(actions.get(0),perception.get(1)));
		primitiveList.add(new PrimitiveInteraction(actions.get(0),perception.get(2)));
		primitiveList.add(new PrimitiveInteraction(actions.get(0),perception.get(3)));
		
		primitiveList.add(new PrimitiveInteraction(actions.get(1),perception.get(0)));
		primitiveList.add(new PrimitiveInteraction(actions.get(2),perception.get(0)));
	}
	
	public int nbInteraction(){
		return primitiveList.size();
	}
	
	public PrimitiveInteraction getInteraction(int i){
		return primitiveList.get(i);
	}
	
	public ArrayList<PrimitiveInteraction> getList(){
		return primitiveList;
	}
		
	/**
	 *  define the valence for each interaction
	 * @param a, p primitive action and perception of an interaction
	 * @return valence of interaction (a,p)
	 */
	public static float valence(PrimitiveAction a, PrimitivePerception p){
		
		float valence=0;
		
		if (a.getId()==0){
			if (p.getId()==0) valence=1;
			if (p.getId()==1) valence=5;
			if (p.getId()==2) valence=-5;
			if (p.getId()==3) valence=200;
			
		}
		else valence=-1;
		
		return valence;
	}
	
	// get the id of the associated primary interaction
	public PrimitiveInteraction getInteractionOfIndex(int id){
		int ret=0;
		if (id>=InteractionList.nbDF){
			ret=id-InteractionList.nbDF;
		}
		else{
			if (id>=0 && id<InteractionList.size*InteractionList.size*4) ret=0;
			else if (id>=  InteractionList.size*InteractionList.size*4 && id<2*InteractionList.size*InteractionList.size*4) ret=1;
			else if (id>=2*InteractionList.size*InteractionList.size*4 && id<3*InteractionList.size*InteractionList.size*4) ret=2;
			else if (id>=3*InteractionList.size*InteractionList.size*4 && id<4*InteractionList.size*InteractionList.size*4) ret=3;
			else if (id>=4*InteractionList.size*InteractionList.size*4 && id<5*InteractionList.size*InteractionList.size*4) ret=4;
			else if (id>=5*InteractionList.size*InteractionList.size*4 && id<6*InteractionList.size*InteractionList.size*4) ret=5;
		}

		return getInteraction(ret);
	}
}
