package version10_2_2_1.Interface;

import version10_2_2_1.environment.Environment;
import version10_2_2_1.platform.Agent;

/**
 * convert signals from the robot into interactions
 * @author simon
 */
public class Perception {

	private Agent agent;

	private int size=4;       						// nb of status
	private boolean[] status;
	
	public Perception(Agent a){
		agent=a;
		
		// "physical" robot input
		status=new boolean[size];
		for (int i=0;i<size;i++){
			status[i]=false;
		}
	}

	
	public void update(){
		
		// get the input vector
		for (int i=0;i<size;i++) status[i]=false;
		if (agent.body.sensors[0]==1) status[0]=true;
		if (agent.body.sensors[0]==2) status[1]=true;
		if (agent.body.sensors[0]==3) status[2]=true;
		if (agent.body.sensors[5]==1) status[3]=true;
	}
	
	
	// interpretation of enacted interaction according to action and inputs
	public PrimitiveInteraction recognize(PrimitiveInteraction intended){
		
		if (intended.getAction().getName().equals(">")){
			if (status[3]) 		return agent.interactionList.getInteraction(3);			// eat
			else if (status[1])return agent.interactionList.getInteraction(2);			// bump
			else if (status[2])return agent.interactionList.getInteraction(1);			// push	
			else 				return agent.interactionList.getInteraction(0);			// forward
		}
		else{
			return intended;
		}
	}
	
	public float[] recognizeEnactedEnsemble(PrimitiveInteraction enacted){

		float[] enactedEnsemble=new float[InteractionList.length];
		
		// initialize enacted ensemble
		for (int i=0;i<InteractionList.length;i++){
			enactedEnsemble[i]=0;
		}
		
		// set primary interaction
		enactedEnsemble[enacted.getIndex()+InteractionList.nbDF]=1;
		
		// set secondary interactions
		for (int i=0;i<InteractionList.size;i++){
			for (int j=0;j<InteractionList.size;j++){
				
				int col=0;
				if (agent.topView.map[i][j].equals(Environment.WALL1)) col=1;
				else if (agent.topView.map[i][j].equals(Environment.FISH1)) col=2;
				else if (agent.topView.map[i][j].equals(Environment.TILE)) col=3;
				else if (agent.topView.map[i][j].equals(Environment.BOX)) col=4;
				
				for (int c=0;c<4;c++) enactedEnsemble[enacted.getIndex()*InteractionList.size*InteractionList.size*4+i*InteractionList.size*4 + j*4 + c]=-1;  	// write failure
				
				if (col>0) enactedEnsemble[enacted.getIndex()*InteractionList.size*InteractionList.size*4+i*InteractionList.size*4 + j*4 + col-1]=1;			// write success
			}
		}
		
		// bias input
		enactedEnsemble[InteractionList.length-1]=1;
		
		return enactedEnsemble;
	}
	
}
