package version10_2_2_1.agnosticMemory;

import java.util.ArrayList;

import version10_2_2_1.Interface.InteractionList;

public class Context {

	public short index_primary;
	public ArrayList<Short> context;
	public ArrayList<Context> included;
	public byte type;
	
	public byte level;
	
	public boolean activated=false;
	public boolean inhibition=false;
	
	public short nb_s=0;
	public short nb_test=0;
	
	public short failure=0;
	
	
	public Context(ArrayList<Short> c, short l, short t){
		index_primary=c.get(c.size()-1);
		context=new ArrayList<Short>();
		level=(byte)l;
		type=(byte)t;
		included=new ArrayList<Context>();
		for (int i=0;i<c.size()-1;i++){
			context.add(c.get(i));
		}
	}
	
	// constructor used to load from a file
	public Context(){
		context=new ArrayList<Short>();
		included=new ArrayList<Context>();
	}
	
	
	public void update(ArrayList<Short> c){
		
		boolean changes=false;
		
		// current context must include the included contexts
		boolean found=false;
		short j=0;
		
		// if all included contexts are included in current context
		if (!found){
			if (index_primary==c.get(c.size()-1)){
				j=0;
				found=false;
				for (short i=0;i<context.size();i++){
					j=0;
					found=false;
					while (!found && j<c.size()-1){
						if (context.get(i).equals(c.get(j))) found=true;
						j++;
					}
					
					if (!found){
						context.remove(i);
						changes=true;
						i--;
					}
				}
			}
			if (changes){
				nb_s=0;
				nb_test=0;
				failure=0;
				included.clear();
			}
		}
	}
	
	public boolean isIncludedIn(ArrayList<Short> c){
		if (this.index_primary!=c.get(c.size()-1)) return false;
		else{
			short i=0;
			short j=0;
			boolean included=true;
			
			while (i<context.size() && included){
				included=false;
				j=0;
				
				while (j<c.size()-1 && !included){
					if (this.context.get(i).equals(c.get(j))) included=true;
					j++;
				}
				i++;
			}		
			return included;
		}
	}
	
	
	public float intersection(ArrayList<Short> c){
		if (this.index_primary!=c.get(c.size()-1)) return 0;
		else{
			short j=0;
			boolean intersect=false;
			float nb=0;
			for (short i=0;i<context.size();i++){
				j=0;
				intersect=false;
				while (j<c.size()-1 && !intersect){
					if (context.get(i).equals(c.get(j))) intersect=true;
					j++;
				}
				if (intersect) nb++;
			}
			return nb/(float)context.size();
		}
	}
	
	public boolean isEqual(Context c){
		if (this.index_primary!=c.index_primary || this.context.size()!=c.context.size()) return false;
		else{
			boolean found=true;
			short i=0;
			while (found && i<context.size()){
				if (!this.context.get(i).equals(c.context.get(i))) found=false;
				i++;
			}
			return found;
		}
	}
	
	// activate context if included in current context
	public void activation(ArrayList<Short> c){
		
		activated=false;
		inhibition=false;
		
		if (this.isIncludedIn(c)){
			activated=true;
			
			// activate including contexts
			for (int i=0;i<included.size();i++){
				included.get(i).activation(c);
			}
			
			// inhibit if an including context is active
			for (short i=0;i<included.size();i++){
				if (included.get(i).activated) inhibition=true;
			}
		}
		else{
			reset();
		}
	}
	
	
	public int[] getImage(){
		int[] ret=new int[InteractionList.length];
		for (short i=0;i<context.size();i++){
			ret[context.get(i)]=255;
		}
		return ret;
	}
	
	public void predictionResult(float output, ArrayList<Short> c){
		if (activated && !inhibition){
			// prediction was right
			if (output*type>0){
				if (nb_test<10) nb_test++;
				if (nb_s<10) nb_s++;
			}
			// prediction was wrong
			else{
				
				updateContext(c);
				
				if (nb_test<10) nb_test++;
				else nb_s=0;
				
				failure++;
			}
		}
		
		for (short i=0;i<included.size();i++){
			included.get(i).predictionResult(output,c);
			
			if (included.get(i).failure>10){
				included.remove(i);
				i--;
			}
		}
	}
	
	public void updateContext(ArrayList<Short> context){
		if (level<2){
			if (context.size()>0){
				// detect if contexts match the current context
				int id=-1;
				float max=0;
				for (short i=0;i<included.size();i++){
					if (!included.get(i).inhibition){
						if (included.get(i).intersection(context)>max){
							max=included.get(i).intersection(context);
							id=i;
						}
					}
				}
					
				if (id!=-1){
					included.get(id).update(context);
					if (this.isEqual(included.get(id))) included.remove(id);
				}
				else included.add(new Context(context,(short)(level+1), (short)-type));
			}
		}
		
		// remove empty contexts
		for (short i=0;i<included.size();i++){
			if (included.get(i).size()<1){
				included.remove(i);
				i--;
			}
		}
	}
	
	
	public short getResult(){
		// if the context is not active, then there are no prediction
		if (!activated) return 0;
		else{
			// if the context is active and not inhibited, then, return the prediction that correspond to the context type
			if (!inhibition) return type;
			else{
				short sum=0;
				for (short i=0;i<included.size();i++){
					sum+=included.get(i).getResult();
				}
				
				if (sum>0) return 1;
				else if (sum<0) return -1;
				else return 0;
			}
		}
	}
	
	public int getUtility(float[] map, byte[] testMap){
		int utility=0;
		
		if (!inhibition){
			//if (nb_s<5 || nb_s<nb_test){
				for (short i=0;i<context.size();i++){
					if (map[context.get(i)]>0){
						utility+= (10-nb_s) * testMap[context.get(i)];
					}
				}
			//}
		}
		else{
			int u=0;
			for (short i=0;i<included.size();i++){
				u=included.get(i).getUtility(map,testMap);
				if (u>utility) utility=u;
			}
		}
		return utility;
	}
	
	////////////////////////////////////////////////////
	
	public int size(){
		return context.size();
	}
	
	public short get(int i){
		return context.get(i);
	}
	
	public void reset(){
		activated=false;
		inhibition=false;
		for (int i=0;i<included.size();i++) included.get(i).reset();
	}
}
