package version10_2_2_1.agnosticMemory;

import version10_2_2_1.Interface.InteractionList;
import version10_2_2_1.Interface.PrimitiveInteraction;
import version10_2_2_1.platform.Agent;

/**
 * Environmental context.
 * This class can also generate a predicted and an improved context
 * @author simon gay
 */
public class EnvironmentContext{
	
	private Agent agent;
	
	public float[][] enactedEnsembles;				// enacted ensembles E_t and E_{t-1}
	public float[][] results;						// results (success, failure, not used for step t and t-1
	
	//public short[][] counter;
	
	public EnvironmentContext(Agent a){
		
		agent = a;
		
		enactedEnsembles=new float[2][InteractionList.length];
		results=new float[2][InteractionList.length];
		//counter=new short[6][InteractionList.length];
		for (int i=0;i<InteractionList.length;i++){
			enactedEnsembles[0][i]=0;
			enactedEnsembles[1][i]=0;
			
			results[0][i]=0;
			results[1][i]=0;
			
			/*for (byte j=0;j<6;j++){
				counter[j][i]=255;
			}*/
		}
	}
	
	
	/**
	 * update the environmental context
	 * @param inter : last enacted primitive interaction
	 */
	public void updateContext(PrimitiveInteraction inter, float[] enacted){
		
		// update timeline
		for (int i=0;i<InteractionList.length;i++){
			enactedEnsembles[1][i]=enactedEnsembles[0][i];
			enactedEnsembles[0][i]=Math.max(0,enacted[i]);
			
			results[1][i]=results[0][i];
			results[0][i]=enacted[i];
		}
		
		// update signatures
		agent.signatureList.learn(enactedEnsembles[1],enactedEnsembles[0] , inter);
		
		// compute predictions
		agent.signatureList.setPredictions(enactedEnsembles[0]);
		
		// set counters
		byte id=0;
		if (inter.getAction().getName().equals(">")){
			if (inter.getPerception().getId()==1) id=1;
			else if (inter.getPerception().getId()==2) id=2;
			else if (inter.getPerception().getId()==3) id=3;
			else id=0;
		}
		else if (inter.getAction().getName().equals("v")) id=4;
		else if (inter.getAction().getName().equals("^")) id=5;
		
		/*for (short i=0;i<InteractionList.length;i++){
			if (enactedEnsembles[1][i]==1 && counter[id][i]>1) counter[id][i]--;
		}*/
		
	}
	
}
