package version10_2_2_1.agnosticMemory;

import java.util.ArrayList;

import version7_2.Interface.InteractionList;

public class Hypothesis{
	public Byte[] path;
	public ArrayList<ArrayList<Short[]>> origin;
	public ArrayList<ArrayList<Short>> prototypes;
	public byte status;
	
	public Hypothesis(Byte[] p, ArrayList<Short[]> o){
		path=p;
		origin=new ArrayList<ArrayList<Short[]>>();
		
		// sort origin paths
		origin=new ArrayList<ArrayList<Short[]>>();
		for (int i=0;i<o.size();i++){
			
			if (origin.size()==0){
				origin.add(new ArrayList<Short[]>());
				origin.get(origin.size()-1).add(o.get(i));
			}
			else{
				int n=0;
				while(n<origin.size() && origin.get(n).get(0)[origin.get(n).get(0).length-1]<o.get(i)[o.get(i).length-1]) n++;
				origin.add(n,new ArrayList<Short[]>());
				origin.get(n).add(o.get(i));
			}
		}
		
		//origin=o;
		status=0;
		prototypes=new ArrayList<ArrayList<Short>>();
		for (int i=0;i<o.size();i++){
			prototypes.add(new ArrayList<Short>());
		}
	}
	
	public void setStatus(boolean s){
		if (s) status=1;
		else status=-1;
	}
	
	public boolean isEqual(Byte[] p, ArrayList<Short[]> o){
		
		if (p.length>=path.length){

			for (int l=0;l<path.length;l++){
				if (path[l]!=p[l]) return false;
			}
			
			for (int i=0;i<o.size();i++){
				
				int j=0;
				boolean found=false;
				while (!found && j<origin.size()){
					
					int l=0;
					found=true;
					while (found && l<origin.get(j).get(0).length){
						if (origin.get(j).get(0)[l]!=o.get(i)[l]) found=false;
						l++;
					}
					j++;
				}
				j=0;
				if (!found) return false;
			}
		}
		else return false;
		
		return true;
	}
	
	
	public byte[] getMap(){
		byte[] map=new byte[InteractionList.length];
		for (int i=0;i<origin.size();i++){
			map[origin.get(i).get(0)[origin.get(i).get(0).length-1]]=1;
		}
		return map;
	}
	
	// detect proto-objects that match the hypothesis
	public void detect(Byte[] p, Short[] origins, SignatureList sigList){
		
		// check path equality
		if (p.length>=path.length){
			boolean identical=true;
			int l=0;
			while (identical && l<path.length){
				if (p[l]!=path[l]) identical=false;
				l++;
			}
			
			if (identical){
				// check origin equality
				boolean found=false;
				int o=0;
				while (!found && o<origin.size()){
				
					l=0;
					found=true;
					while (found && l<origin.get(o).get(0).length){
						if (origins[l]!=origin.get(o).get(0)[l]) found=false;
						l++;
					}
					o++;
				}
				
				if (found){
					
					short val=origins[origins.length-1];
					found=false;
					int i=0;
					while (!found && i<prototypes.get(o-1).size()){
						if (prototypes.get(o-1).get(i).equals(val)) found=true;
						i++;
					}
					if (!found){
						
						if (p.length<=path.length) prototypes.get(o-1).add(val);
						else{
							// check for the validity of the path
							int sig_index=origin.get(o-1).get(0)[origin.get(o-1).get(0).length-1];
							Byte[] path_temp=new Byte[p.length-path.length-1];
							for (int l2=0;l2<path_temp.length;l2++){
								path_temp[l2]=p[l2+path.length+1];
							}
							//System.out.print(" ~~~~~~~~~~ "+sig_index+" ; [");
							//for (int l2=p.length-1;l2>=0;l2--) System.out.print(p[l2]+", ");
							//System.out.print("] -> {");
							//for (int l2=path_temp.length-1;l2>=0;l2--) System.out.print(path_temp[l2]+", ");
							//System.out.print("} ->");
							
							short[] map=new short[InteractionList.length];
							found=(new Projection(sigList.signatureList.get(sig_index), path_temp, sigList.signatureList, (byte)-1)).getResult(map);
							//System.out.print(found+" ; {");
							//for (int j=0;j<InteractionList.length-1;j++){
							//	if (map[j]!=0) System.out.print(j+", ");
							//}
							//System.out.println("} ");
							if (found) prototypes.get(o-1).add(val);
						}
					}
				}
			}
		}
	}
	

	// detect if a same element is recognized as two different proto-objects
	public boolean checkValidity(){
		boolean valide=true;
		
		int p1=0;
		int p2=0;
		int o1=0;
		int o2=0;
		
		for (p1=0;p1<prototypes.size();p1++){
			
			for (o1=0;o1<prototypes.get(p1).size();o1++){
				
				boolean found=false;
				
				for (p2=p1+1;p2<prototypes.size();p2++){
					
					for (o2=0;o2<prototypes.get(p2).size();o2++){
						
						if (prototypes.get(p1).get(o1).equals(prototypes.get(p2).get(o2))){
							found=true;
							prototypes.get(p2).remove(o2);
							o2--;
						}	
					}
				}
				if (found){
					prototypes.get(p1).remove(o1);
					o1--;
				}	
			}	
		}
		return valide;
	}
	
	public boolean isActive(){

		boolean ret=true;
		int i=0;
		while (ret && i<prototypes.size()){
			if (prototypes.get(i).size()==0) ret=false;
			i++;
		}
		return ret;
	}
	
	
	// reset the list of detected proto-objects
	public void resetDetection(){
		for (int i=0;i<prototypes.size();i++) prototypes.get(i).clear();
	}
	
	// display the hypothesis
	public void print(){
		
		System.out.print("[");
		for (int l=path.length-1;l>=0;l--) System.out.print(path[l]+", ");
		System.out.println("]");
		
		for (int k=0;k<origin.size();k++){
			System.out.print("--- (");
			for (int l=origin.get(k).get(0).length-1;l>=0;l--) System.out.print(origin.get(k).get(0)[l]+", ");
			

			System.out.println(" )");
			System.out.print("     -> {");
			for (int p=0;p<prototypes.get(k).size();p++){
				System.out.print(prototypes.get(k).get(p)+", ");
			}
			System.out.println(" }");
		}
		//System.out.println("--------------------------------------");
	}
}

