package version10_2_2_1.agnosticMemory;

import java.util.ArrayList;

public class HypothesisMemory {

	public ArrayList<Hypothesis> memory;
	public ArrayList<ArrayList<Short[]>> groups;
	
	public HypothesisMemory(){
		memory=new ArrayList<Hypothesis>();
		groups=new ArrayList<ArrayList<Short[]>>();
	}
	
	public void addHypothesis(Byte[] p, ArrayList<Short[]> o){
		
		// check for existing hypothesis
		boolean found=false;
		int mem=0;
		while (!found && mem<memory.size()){
			if (memory.get(mem).isEqual(p, o)) found=true;
			mem++;
		}
		
		// if not known, create new hypothesis
		if (!found){
		
			int div=0;
			
			//for (int i=0;i<o.size();i++){
			//	System.out.print(" ---- [");
			//	for (int j=0;j<o.get(i).length;j++) System.out.print(o.get(i)[j]+", ");
			//	System.out.println("]");
			//}
	
			// get the index of the first divergence
			boolean identical=true;
			div=0;
			while (identical && div<p.length+1){
				for (int j=0;j<o.size();j++){
					for (int k=j;k<o.size();k++){
						if (o.get(j)[div]!=o.get(k)[div]) identical=false;
					}
				}
				div++;
			}
	
			// remove redundant origin paths
			for (int j=0;j<o.size();j++){
				for (int k=j+1;k<o.size();k++){
					if (o.get(j)[div-1]==o.get(k)[div-1]){
						o.remove(k);
						k--;
					}
				}
			}

			if (o.size()>1){
				Byte[] path=new Byte[div-1];
				ArrayList<Short[]> origin=new ArrayList<Short[]>();
				
				for (int l=0;l<div-1;l++){
					path[l]=p[l];
				}
				
				for (int j=0;j<o.size();j++){
					Short[] temp=new Short[div];
					for (int l=0;l<div;l++){
						temp[l]=o.get(j)[l];
					}
					origin.add(temp);
				}
				
				memory.add(new Hypothesis(path, origin));
			}
		}
	}
	
	public void print(){
		for (int i=0;i<memory.size();i++){
			memory.get(i).print();
		}
	}

	// check if the hypothesis is valid (strict separation between proto-objects)
	/*public void checkValidity(){
		for (int i=0;i<memory.size();i++){
			//memory.get(i).checkValidity();
		}
	}*/
	
	// remove hypothesis that are included in others
	public void removeRedundant(){
		for (int i=0;i<memory.size();i++){
			for (int j=0;j<memory.size();j++){
				
				if (memory.get(i).origin.size()<memory.get(j).origin.size()){
					boolean found=true;
					int o1=0;
					int o2=0;
					while (found && o1<memory.get(i).origin.size()){
						o2=0;
						found=false;
						
						while (!found && o2<memory.get(j).origin.size()){
							found=(memory.get(i).origin.get(o1).get(0).length==memory.get(j).origin.get(o2).get(0).length);
							int l=0;
							while (found && l<memory.get(i).origin.get(o1).get(0).length){
								if (memory.get(i).origin.get(o1).get(0)[l].equals(memory.get(j).origin.get(o2).get(0)[l])) found=false;
								l++;
							}
							o2++;
						}
						o1++;
					}
					
					if (found){
						memory.remove(j);
						if (i<j) j--;
						else{
							i--;
							j--;
						}
					}
				}
			}
		}
	}
	
	public ArrayList<Hypothesis> detectActive(int id, SignatureList sigList){
		
		ArrayList<Hypothesis> ret=new ArrayList<Hypothesis>();
		
		for (int i=0;i<memory.size();i++){
			if (memory.get(i).isActive()){
				// get elements of the constructible that cannot be moved
				ArrayList<ArrayList<Short>> prototypes=new ArrayList<ArrayList<Short>>();
				
				for (int p=0;p<memory.get(i).prototypes.size();p++){
					
					boolean found=false;
					int e=0;
					while (!found && e<memory.get(i).prototypes.get(p).size()){
						
						int k=0;
						while (!found && k<4){
							if (k!=id){ // don't try to construct an object that is already present
								
								int m=0;
								while (!found && m<sigList.workMemory[k].movableInstances.size()){
									if (sigList.workMemory[k].movablePattern.get(m).contains(memory.get(i).prototypes.get(p).get(e)))
										found=true;
									m++;
								}
							}
							k++;
						}
						e++;
					}
					
					if (!found){
						prototypes.add(new ArrayList<Short>());
						for (int m=0;m<memory.get(i).prototypes.get(p).size();m++)
							prototypes.get(prototypes.size()-1).add(memory.get(i).prototypes.get(p).get(m));
					}
				}
				
				//memory.get(i).print();
				//System.out.println(" non-movable :");
				//for (int p=0;p<prototypes.size();p++){
				//	for (int j=0;j<prototypes.get(p).size();j++) System.out.print(prototypes.get(p).get(j)+", ");
				//	System.out.println();
				//}
				
				// case where one object max cannot be moved
				if (prototypes.size()<=1){
					ret.add(memory.get(i));
				}
				// other case : it must exist a common path in each non-movable group
				else{
					// TODO
				}
			}
		}
		return ret;
	}
	
}
