package version10_2_2_1.agnosticMemory;

import java.util.ArrayList;

import version10_2_2_1.Interface.InteractionList;


// an equation defines the boolean equation of an image of an interaction
// an equation is under the form of :
// an interaction if type = 0
// an equation of type   node1 + node2 + ... + nodeN   if interaction false by default
// an equation of type /(node1 + node2 + ... + nodeN)  if interaction true by default
public class Projection{

	public static short LENGTH=7;
	
	public static int RELIABILITY=5;
	
	public byte level;
	public byte type;			// 0 -> leaf, 1 -> equation default true, -1 -> equation default false
	public Signature leaf;
	public ArrayList<Node> equation_t;
	public ArrayList<Node> equation_f;
	
	// first level
	public Projection(Signature s, Byte[] path, ArrayList<Signature> sig, byte previous){
		level=(byte)(path.length-1);
		type=0;
		leaf=s;

		if (level>=0){
		
			// define new type
			if (leaf.nb_s_t==100) type=1;
			if (leaf.nb_s_f==100) type=-1;
			
			if (type!=0){
				equation_t=new ArrayList<Node>();
				equation_f=new ArrayList<Node>();
				
				projection(sig, path, previous);
			}
		}
	}
		
	private Projection(Signature s, Byte[] path, byte l, ArrayList<Signature> sig, byte previous){
		level=l;
		type=0;
		leaf=s;
		
		// projection only if still in the range
		if (level>=0){
			
			// define new type
			if (leaf.nb_s_t==100) type=1;
			if (leaf.nb_s_f==100) type=-1;
			
			if (type!=0){
				equation_t=new ArrayList<Node>();
				equation_f=new ArrayList<Node>();

				projection(sig, path, previous);
			}
		}
	}
		
	// replace leaf elements of the equation by their  own equations
	private void projection(ArrayList<Signature> sig, Byte[] path, byte previous){

		for (int c=0;c<leaf.context_t.size();c++){
			// if the context is related to the primary interaction mov and is reliable, add new node to characterize this context
			if (leaf.context_t.get(c).nb_s>=RELIABILITY && leaf.context_t.get(c).nb_s==leaf.context_t.get(c).nb_test){	// reliable
				int index=leaf.context_t.get(c).index_primary-InteractionList.nbDF;
				
				if (index==path[path.length-1-level])
					equation_t.add(new Node(leaf.context_t.get(c), sig, path,(byte)(level-1), (byte)index));
			}
		}
		
		for (int c=0;c<leaf.context_f.size();c++){
			// if the context is related to the primary interaction mov and is reliable, add new node to characterize this context
			if (leaf.context_f.get(c).nb_s>=RELIABILITY && leaf.context_f.get(c).nb_s==leaf.context_f.get(c).nb_test){	// reliable
				int index=leaf.context_f.get(c).index_primary-InteractionList.nbDF;
				
				if (index==path[path.length-1-level])
					equation_f.add(new Node(leaf.context_f.get(c), sig, path, (byte)(level-1), (byte)index));
			}
		}
	}
		
		
	// define if the subcontext is active according to its equation. Get the pattern of the object in 'pattern'
	public boolean getResult(short[] pattern){
		
		if (level==-1){
			if (leaf.getPrediction()==1){
				for (int c=0;c<leaf.context_f.size();c++){
					for (int i=0;i<leaf.context_f.get(c).size();i++){
						pattern[leaf.context_f.get(c).get(i)]=1;
					}
				}
				
				return true;
			}
			else return false;
		}
		else if (type==0){
			return false;				// end of the projection
		}
		else{
			boolean default_t=(leaf.nb_s_t==100);		// case default true 
			boolean default_f=(leaf.nb_s_f==100);		// case default false
			
			boolean ret_t=false;
			boolean ret_f=false;
			
			int i=0;
			while (!ret_t && i<equation_f.size()){
				if (equation_f.get(i).getResult(pattern)) ret_t=true;
				i++;
			}
			i=0;
			while (!ret_f && i<equation_t.size()){
				if (equation_t.get(i).getResult(pattern)) ret_f=true;
				i++;
			}
			return (default_t && !ret_f) || (default_f && ret_t);
		}
	}
	
	//-------------------------------------------------------------------------------------------
	// a node is a primitive context with its included contexts
	// a context is under the form of an equation eq && /included
	private class Node{
		public ArrayList<Projection> eq;			// interactions that compose the context
		private ArrayList<Node> included;		// included contexts of this context
		
		public Node(Context c, ArrayList<Signature> sig, Byte[] path, byte l, byte previous){
			
			eq=new ArrayList<Projection>();
			included=new ArrayList<Node>();
			
			// add elements of the context
			for (short i=0;i<c.size();i++){
				eq.add(new Projection(sig.get(c.context.get(i)), path, l, sig, previous) );
			}
			
			// add included contexts
			for (short i=0;i<c.included.size();i++){
				included.add(new Node(c.included.get(i),sig,path, l, previous));
			}
		}
		
		public boolean getResult(short[] pattern){
			
			short i=0;
			while (i<eq.size()){
				if (!eq.get(i).getResult(pattern)) return false;
				i++;
			}
			
			i=0;
			while (i<included.size()){
				if (included.get(i).getResult(pattern)) return false;
				i++;
			}	
			return true;
		}

	}
	
}
