package version10_2_2_1.agnosticMemory;

import java.util.ArrayList;

import version10_2_2_1.Interface.InteractionList;
import version10_2_2_1.Interface.PrimitiveInteraction;

/**
 * signature of an interaction
 * @author simon gay
 */
public class Signature {

	public PrimitiveInteraction interaction;	// primary interaction or associated interaction used to define selected action 
	
	private byte prediction;
	
	public ArrayList<Context> context_t;
	public ArrayList<Context> context_f;

	// "default" context t and f
	public boolean activated_t=false;
	public boolean inhibition_t=false;
	public byte nb_s_t=0;
	public byte nb_test_t=0;
	
	public boolean activated_f=false;
	public boolean inhibition_f=false;
	public byte nb_s_f=0;
	public byte nb_test_f=0;
	
	public byte[] testMap;
	
	public short index;
	
	public boolean reliable=false;
	
	
	public Signature(PrimitiveInteraction inter, short id){
		interaction=inter;
		
		index=id;

		prediction=0;
		
		context_t=new ArrayList<Context>();
		context_f=new ArrayList<Context>();
		
		testMap=new byte[InteractionList.length];
		for (short i=0;i<InteractionList.length;i++) testMap[i]=20;
	}
	
	/**
	 * predict the result of an interaction in a given context
	 * @param img : environmental context
	 * @return the certitude of success (1->absolute certitude of success, -1-> absolute certitude of failure 
	 */
	public void prediction(float[] img){
		float sum=0;

		/////////////////////////////////////////////////
		// define previous environmental context
		ArrayList<Short> context=new ArrayList<Short>();
		for (short i=0;i<img.length-1;i++){
			if (img[i]>0) context.add(i);
		}
		
		///////////////////////////////////////////////////////////////////
		// reset activation and inhibition
		activated_t=false;
		inhibition_t=false;
		activated_f=false;
		inhibition_f=false;
		
		///////////////////////////////////////////////////////////////////
		// activate contexts that match the current context
		
		// default contexts
		activated_t=(nb_s_f<100);
		activated_f=(nb_s_t<100);
		
		// other contexts
		for (short i=0;i<context_t.size();i++) context_t.get(i).activation(context);
		for (short i=0;i<context_f.size();i++) context_f.get(i).activation(context);
		
		//////////////////////////////////////////////////////////////////
		// inhibition of default contexts
		for (short i=0;i<context_t.size();i++){
			if (context_t.get(i).activated) inhibition_t=true;
		}
		for (short i=0;i<context_f.size();i++){
			if (context_f.get(i).activated) inhibition_f=true;
		}
		
		///////////////////////////////////////////////////////////////////
		// define prediction
		int res_t=0;
		int res_f=0;
		
		if (activated_t){
			// if the context is active and not inhibited, then, return the prediction that correspond to the context type
			if (!inhibition_t) res_t=1;
			else{
				sum=0;
				for (short i=0;i<context_t.size();i++){
					sum+=context_t.get(i).getResult();
				}
				
				if (sum>0) res_t=1;
				else if (sum<0) res_t=-1;
				else res_t=0;
			}
		}
		
		if (activated_f){
			// if the context is active and not inhibited, then, return the prediction that correspond to the context type
			if (!inhibition_f) res_f=-1;
			else{
				sum=0;
				for (short i=0;i<context_f.size();i++){
					sum+=context_f.get(i).getResult();
				}
				
				if (sum>0) res_f=1;
				else if (sum<0) res_f=-1;
				else res_f=0;
			}
		}
		
		if (res_t>0 && res_f>=0) prediction=1;
		else if (res_t>=0 && res_f>0) prediction=1;
		else if (res_t<0 && res_f<=0) prediction=-1;
		else if (res_t<=0 && res_f<0) prediction=-1;
		else prediction=0;
	}
	
	/**
	 * integrate the result of enaction cycle
	 * @param img : environmental context E_{t-1} 
	 * @param output result of enaction (success->1, failure->-1)
	 */
	public void learn(float[] img, float output, ArrayList<Signature> signatureList){
	
		/////////////////////////////////////////////////
		// integrate learning context
		for (short i=0;i<InteractionList.length;i++){
			if (img[i]>0 && testMap[i]>1) testMap[i]-=10;
			else if (testMap[i]<20) testMap[i]++;
		}
		
		/////////////////////////////////////////////////
		// define previous context
		ArrayList<Short> context=new ArrayList<Short>();
		for (short i=0;i<img.length-1;i++){
			if (img[i]>0) context.add(i);
		}
		
		/////////////////////////////////////////////////
		// integrate prediction results
		if (activated_t && !inhibition_t){
			if (output>0){
				if (nb_test_t<100) nb_test_t++;
				if (nb_s_t<100)    nb_s_t++;
			}
			else{
				updateContext_t(context);
				
				if (nb_test_t<100) nb_test_t++;
				else nb_s_t=0;
			}
		}
		
		if (activated_f && !inhibition_f){
			if (output<0){
				if (nb_test_f<100) nb_test_f++;
				if (nb_s_f<100)    nb_s_f++;
			}
			else{
				updateContext_f(context);
				
				if (nb_test_f<10) nb_test_f++;
				else nb_s_f=0;
			}
		}
		

		for (short i=0;i<context_t.size();i++){
			context_t.get(i).predictionResult(output,context);
			
			if (context_t.get(i).failure>10){
				context_t.remove(i);
				i--;
			}
		}
		
		for (short i=0;i<context_f.size();i++){
			context_f.get(i).predictionResult(output,context);
			
			if (context_f.get(i).failure>10){
				context_f.remove(i);
				i--;
			}
		}

		/////////////////////////////////////////////////
		// remove redundant contexts
		for (short i=0;i<context_t.size();i++){
			for (short j=(short) (i+1);j<context_t.size();j++){
				if (context_t.get(i).isEqual(context_t.get(j))){
					context_t.remove(j);
					j--;
				}
			}
		}
		
		for (short i=0;i<context_f.size();i++){
			for (short j=(short) (i+1);j<context_f.size();j++){
				if (context_f.get(i).isEqual(context_f.get(j))){
					context_f.remove(j);
					j--;
				}
			}
		}
		
		/////////////////////////////////////////////////
		// remove contexts with errors but no included
		for (short i=0;i<context_t.size();i++){
			if (context_t.get(i).failure>0 && context_t.get(i).included.size()==0){
				context_t.remove(i);
				i--;
			}
		}
		
		for (short i=0;i<context_f.size();i++){
			if (context_f.get(i).failure>0 && context_f.get(i).included.size()==0){
				context_f.remove(i);
				i--;
			}
		}
		
		/////////////////////////////////////////////////
		// remove contexts when default context is sufficient
		if (nb_s_t==100) context_f.clear();
		if (nb_s_f==100) context_t.clear();
	}
	
	
	public float getPrediction(){
		return prediction;
	}
	
	////////////////////////////////////////////////////////////////////////////////////////////////////////
	public int getLearningUtility(float[] map){
		int utility=0;
		reliable=true;
		
		int u=0;
		//if (prediction>=0){
			if (nb_s_t<100 && nb_s_f<100){
				int sum=0;
				for (short i=0;i<map.length;i++){
					if (map[i]>0) sum+= (10-(nb_s_t/10)) * testMap[i];
				}
				utility=sum;
				reliable=false;
			}
			for (short i=0;i<context_t.size();i++){
				u=context_t.get(i).getUtility(map,testMap);
				u=u*5;
				if (u>utility) utility=u;
			}
		//}
		
		u=0;
		//if (prediction<=0){
			if (nb_s_t<100 && nb_s_f<100){
				int sum=0;
				for (short i=0;i<map.length;i++){
					if (map[i]>0) sum+= (10-(nb_s_f/10)) * testMap[i];
				}
				if (sum>utility) utility=sum;
			}
			for (short i=0;i<context_f.size();i++){
				u=context_f.get(i).getUtility(map,testMap);
				u=u*5;
				if (u>utility) utility=u;
			}
		//}
		return utility;
	}
	
	/*public float[] getCuriosityUtility(ArrayList<Byte[]> pathList){
		float[] ret=new float[6];
		int[] dist=new int[6];
		for (int k=0;k<6;k++){
			ret[k]=0;
			dist[k]=Projection.LENGTH+1;
		}
		
		// define the shortest paths that end with a same interaction
		for (int p=0;p<pathList.size();p++){
			if (pathList.get(p).length<dist[pathList.get(p)[0]]) dist[pathList.get(p)[0]]=pathList.get(p).length;
		}
		
		
		if (nb_s_f<100){
			if (nb_s_t<100){
				for (int p=0;p<pathList.size();p++){
					if (pathList.get(p).length==dist[pathList.get(p)[0]]){
						ret[pathList.get(p)[pathList.get(p).length-1]]+=(100-nb_s_t) * (float)Math.exp(-0.1*pathList.get(p).length)/100;
					}
				}
			}
			
			for (int c=0;c<context_t.size();c++){
				
				int index_previous=context_t.get(c).index_primary-InteractionList.nbDF;
				for (int p=0;p<pathList.size();p++){
					if (pathList.get(p)[0]==index_previous && pathList.get(p).length==dist[pathList.get(p)[0]]){
						ret[pathList.get(p)[pathList.get(p).length-1]]+=(10-context_t.get(c).nb_s) * (float)Math.exp(-0.5*pathList.get(p).length) /100;
					}
				}
			}
		}
		
		if (nb_s_t<100){
			if (nb_s_f<100){
				for (int p=0;p<pathList.size();p++){
					if (pathList.get(p).length==dist[pathList.get(p)[0]]){
						ret[pathList.get(p)[pathList.get(p).length-1]]+=(100-nb_s_f) * (float)Math.exp(-0.1*pathList.get(p).length) /100;
					}
				}
			}
			
			for (int c=0;c<context_f.size();c++){
				
				int index_previous=context_f.get(c).index_primary-InteractionList.nbDF;
				for (int p=0;p<pathList.size();p++){
					if (pathList.get(p)[0]==index_previous && pathList.get(p).length==dist[pathList.get(p)[0]]){
						ret[pathList.get(p)[pathList.get(p).length-1]]+=(10-context_f.get(c).nb_s) * (float)Math.exp(-0.5*pathList.get(p).length) /100;
					}
				}
			}
		}

		return ret;
	}*/
	
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////
	public PrimitiveInteraction interaction(){
		return interaction;
	}
	
	
	public void updateContext_t(ArrayList<Short> context){
		if (context.size()>0){
			// detect if contexts match the current context
			int id=-1;
			float max=0;
			for (short i=0;i<context_t.size();i++){
				if (!context_t.get(i).inhibition){
					if (context_t.get(i).intersection(context)>max){
						max=context_t.get(i).intersection(context);
						id=i;
					}
				}
			}
				
			if (id!=-1) context_t.get(id).update(context);
			else{
				context_t.add(new Context(context,(short) 0, (short)-1));
				for (short i=0;i<InteractionList.length;i++) testMap[i]=20; 
			}
		}
	}
	
	public void updateContext_f(ArrayList<Short> context){
		if (context.size()>0){
			// detect if contexts match the current context
			int id=-1;
			float max=0;
			for (short i=0;i<context_f.size();i++){
				if (!context_f.get(i).inhibition){
					if (context_f.get(i).intersection(context)>max){
						max=context_f.get(i).intersection(context);
						id=i;
					}
				}
			}
				
			if (id!=-1) context_f.get(id).update(context);
			else context_f.add(new Context(context,(short) 0, (short) 1));
		}
	}
	
	public void reset(){
		activated_t=false;
		inhibition_t=false;
		activated_f=false;
		inhibition_f=false;
		
		for (int i=0;i<context_t.size();i++) context_t.get(i).reset();
		for (int i=0;i<context_f.size();i++) context_f.get(i).reset();
	}
	
	////////////////////////////////////////////////////////
	// fast exp function approximation
	public static double exp(double val) {
	    final long tmp = (long) (1512775 * val + (1072693248 - 60801));
	    return Double.longBitsToDouble(tmp << 32);
	}
	
}
