package version10_2_2_1.display;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import version10_2_2_1.agnosticMemory.Projection;
import version10_2_2_1.environment.Environment;
import version10_2_2_1.platform.Agent;
import version10_2_2_1.Interface.InteractionList;

/**
 * Frame to display images of signatures
 * @author simon
 */

/* - inherited from EnvPanel :
 *   Agent agent      : pointer to the agent
 */
public class InstancePanel extends EnvPanel implements MouseListener{

	private static final long serialVersionUID = 1L;
	
	private int clic_x=0;
	private int clic_y=0;
	
	private short[] pattern;
	
	private boolean[][] selection;
	private boolean[] distance;
	
	private int selected_path=-1;
	private boolean change=false;
	
	private byte type=0;				// type of list to display
	private byte hypothesis=0;
	
	private byte[] map;
	
	public InstancePanel(Agent a){
		super(a);
		addMouseListener(this);
		
		pattern=new short[InteractionList.length];
		for (int i=0;i<InteractionList.length;i++){
			pattern[i]=0;
		}
		
		selection=new boolean[2][5];
		distance=new boolean[Projection.LENGTH+1];
		
		map=new byte[InteractionList.length];
	}
	
	public void setAgent(Agent a){
		agent=a;
	}
	
	public void paintComponent(Graphics g){
		g.setColor(Color.white);
		
		g.fillRect(0, 0, 1500, 1000);
		
		///////////////////////////////////////////////////////////////////////////////////////
		// draw center of object instances
		
		// draw environment
		for (int i=0;i<agent.getMain().env.getWidth();i++){
			for (int j=0;j<agent.getMain().env.getHeight();j++){
				
				if (agent.getMain().env.isWall(i, j)) g.setColor(Environment.WALL1);
				else if (agent.getMain().env.isFood(i, j)) g.setColor(Environment.FISH1);
				else if (agent.getMain().env.isTile(i, j)) g.setColor(Environment.TILE);
				else g.setColor(Environment.FIELD_COLOR);
				
				g.fillRect(40*i, 40*(agent.getMain().env.getHeight()-1-j), 40, 40);
			}	
		}
		
		// add boxes
		g.setColor(Environment.BOX);
		for (short i=0;i<agent.getMain().env.boxList.size();i++){
			g.fillRect(40*(int)agent.getMain().env.boxList.get(i).getpx(), 40*(int)(agent.getMain().env.getHeight()-1-agent.getMain().env.boxList.get(i).getpy()), 40, 40);
		}
		
		// draw agent's position and orientation
		g.setColor(Color.gray);
		g.fillOval((int)(40*agent.body.getpx()+5), (int)(40*(agent.getMain().env.getHeight()-1-agent.body.getpy())+5), 30, 30);
		
		float dx= 20*(float)Math.cos(agent.body.getrz());
		float dy= 20*(float)Math.sin(-agent.body.getrz());
		g.drawLine((int)(40*agent.body.getpx()+20),    (int)(40*(agent.getMain().env.getHeight()-1-agent.body.getpy())+20),
				   (int)(40*agent.body.getpx()+20+dx), (int)(40*(agent.getMain().env.getHeight()-1-agent.body.getpy())+20+dy));
		
		////////////////////////////////////////////////////////////////////////
		//draw center of object instances (in external observer's point of view)
		if (type==0)
		for (short k=0;k<4;k++){
			if (selection[0][k+1]){
				
				// define the color
				if (k==0) g.setColor(Color.red);
				else if (k==1) g.setColor(new Color(150,150,0));
				else if (k==2) g.setColor(Color.green);
				else if (k==3) g.setColor(Color.blue);
				else g.setColor(Color.black);
				
				for (int p=0;p<agent.signatureList.workMemory[k].instanceList.size();p++){
					
					if (distance[agent.signatureList.workMemory[k].instanceList.get(p).length]){
						
						// compute position of the instance
						float x=agent.body.getpx();
						float y=agent.body.getpy();
						float rz=agent.body.getrz();
						
						int i=agent.signatureList.workMemory[k].instanceList.get(p).length-1;
						while (i>=0){
							
							// move forward (>O, >P and >F)
							if (agent.signatureList.workMemory[k].instanceList.get(p)[i]<=1 || agent.signatureList.workMemory[k].instanceList.get(p)[i]==3){
								// move forward
								dx= 1*(float)Math.cos(rz);
								dy= 1*(float)Math.sin(rz);
								
								x+=dx;
								y+=dy;
							}
							else if (agent.signatureList.workMemory[k].instanceList.get(p)[i]==4) rz-=Math.PI/2;
							else if (agent.signatureList.workMemory[k].instanceList.get(p)[i]==5) rz+=Math.PI/2;

							i--;
						}
						
						// draw instance position
						g.drawOval((int)(40*x+10+k), (int)(40*(agent.getMain().env.getHeight()-1-y))+10+k, 20, 20);
			
						dx= 15*(float)Math.cos(rz);
						dy= 15*(float)Math.sin(-rz);
						g.drawLine((int)(40*x+20+k),    (int)(40*(agent.getMain().env.getHeight()-1-y)+20+k),
								   (int)(40*x+20+dx+k), (int)(40*(agent.getMain().env.getHeight()-1-y)+20+dy+k));
					}
				}
			}
		}
		

		// draw movable instances
		if (type==1){
			for (int k=0;k<4;k++){
				
				if (selection[0][k+1]){

					// define the color
					if (k==0) g.setColor(Color.red);
					else if (k==1) g.setColor(new Color(150,150,0));
					else if (k==2) g.setColor(Color.green);
					else if (k==3) g.setColor(Color.blue);
					else g.setColor(Color.black);

					for (int p=0;p<agent.signatureList.workMemory[k].movableInstances.size();p++){
							
						// compute position of the instance
						float x=agent.body.getpx();
						float y=agent.body.getpy();
						float rz=agent.body.getrz();
						
						int i=agent.signatureList.workMemory[k].movableInstances.get(p).get(0).length-1;
						while (i>=0){

							// move forward (>O, >P and >F)
							if (agent.signatureList.workMemory[k].movableInstances.get(p).get(0)[i]<=1 ||
								agent.signatureList.workMemory[k].movableInstances.get(p).get(0)[i]==3){
								// move forward
								dx= 1*(float)Math.cos(rz);
								dy= 1*(float)Math.sin(rz);
								
								x+=dx;
								y+=dy;
							}
							else if (agent.signatureList.workMemory[k].movableInstances.get(p).get(0)[i]==4) rz-=Math.PI/2;
							else if (agent.signatureList.workMemory[k].movableInstances.get(p).get(0)[i]==5) rz+=Math.PI/2;
							
							i--;
						}
						
						
						// last interaction of the second path
						float x2=x;
						float y2=y;
						float rz2=rz;
						
						// move forward (>O, >P and >F)
						if (agent.signatureList.workMemory[k].movableInstances.get(p).get(1)[0]<=1 || agent.signatureList.workMemory[k].movableInstances.get(p).get(1)[0]==3){
							// move forward
							dx= 1*(float)Math.cos(rz);
							dy= 1*(float)Math.sin(rz);
							
							x2+=dx;
							y2+=dy;
						}
						else if (agent.signatureList.workMemory[k].movableInstances.get(p).get(1)[0]==4) rz2-=Math.PI/2;
						else if (agent.signatureList.workMemory[k].movableInstances.get(p).get(1)[0]==5) rz2+=Math.PI/2;

						
						// draw instance position
						g.fillOval((int)(40*x+17), (int)(40*(agent.getMain().env.getHeight()-1-y))+17, 6, 6);
						g.drawLine((int)(40*x+20),  (int)(40*(agent.getMain().env.getHeight()-1-y)+20),
								   (int)(40*x2+20), (int)(40*(agent.getMain().env.getHeight()-1-y2)+20));
						
						//second part
						dx= 15*(float)Math.cos(rz2);
						dy= 15*(float)Math.sin(-rz2);
						g.drawOval((int)(40*x2+14), (int)(40*(agent.getMain().env.getHeight()-1-y2))+14, 12, 12);
						g.drawLine((int)(40*x2+20),   (int)(40*(agent.getMain().env.getHeight()-1-y2)+20),
								   (int)(40*x2+20+dx),(int)(40*(agent.getMain().env.getHeight()-1-y2)+20+dy));
					}
				}
			}
		}
		
		// draw final list of movable instances
		if (type==2)
			for (int k=0;k<4;k++){
				
				if (selection[0][k+1]){

					// define the color
					if (k==0) g.setColor(Color.red);
					else if (k==1) g.setColor(new Color(150,150,0));
					else if (k==2) g.setColor(Color.green);
					else if (k==3) g.setColor(Color.blue);
					else g.setColor(Color.black);

					for (int p=0;p<agent.signatureList.workMemory[k].movables.size();p++){
							
						// compute position of the instance
						float x=agent.body.getpx();
						float y=agent.body.getpy();
						float rz=agent.body.getrz();
						
						int i=agent.signatureList.workMemory[k].movables.get(p).length-1;
						while (i>=0){

							// move forward (>O, >P and >F)
							if (agent.signatureList.workMemory[k].movables.get(p)[i]<=1 ||
								agent.signatureList.workMemory[k].movables.get(p)[i]==3){
								// move forward
								dx= 1*(float)Math.cos(rz);
								dy= 1*(float)Math.sin(rz);
								
								x+=dx;
								y+=dy;
							}
							else if (agent.signatureList.workMemory[k].movables.get(p)[i]==4) rz-=Math.PI/2;
							else if (agent.signatureList.workMemory[k].movables.get(p)[i]==5) rz+=Math.PI/2;
							
							i--;
						}
						
						
						// last interaction of the second path
						float x2=x;
						float y2=y;
						float rz2=rz;
						
						// move forward (>O, >P and >F)
						if (k<=1 || k==3){
							// move forward
							dx= 1*(float)Math.cos(rz);
							dy= 1*(float)Math.sin(rz);
							
							x2+=dx;
							y2+=dy;
						}
						else if (k==4) rz2-=Math.PI/2;
						else if (k==5) rz2+=Math.PI/2;

						
						// draw instance position
						g.fillOval((int)(40*x+17), (int)(40*(agent.getMain().env.getHeight()-1-y))+17, 6, 6);
						g.drawLine((int)(40*x+20),  (int)(40*(agent.getMain().env.getHeight()-1-y)+20),
								   (int)(40*x2+20), (int)(40*(agent.getMain().env.getHeight()-1-y2)+20));
						
						//second part
						dx= 15*(float)Math.cos(rz2);
						dy= 15*(float)Math.sin(-rz2);
						g.drawOval((int)(40*x2+14), (int)(40*(agent.getMain().env.getHeight()-1-y2))+14, 12, 12);
						g.drawLine((int)(40*x2+20),   (int)(40*(agent.getMain().env.getHeight()-1-y2)+20),
								   (int)(40*x2+20+dx),(int)(40*(agent.getMain().env.getHeight()-1-y2)+20+dy));
					}
				}
			}
		
		///////////////////////////////////////////////////////////////////////
		// draw detected proto-objects
		//if (type==0)
		for (short k=0;k<4;k++){
			
			if (selection[1][k+1]){
				
				// define the color
				if (k==0) g.setColor(Color.red);
				else if (k==1) g.setColor(new Color(150,150,0));
				else if (k==2) g.setColor(Color.green);
				else if (k==3) g.setColor(Color.blue);
				else g.setColor(Color.black);
				
				for (int p=0;p<agent.signatureList.workMemory[k].pathList.size();p++){
					for (int p2=0;p2<agent.signatureList.workMemory[k].pathList.get(p).size();p2++){
					if (distance[agent.signatureList.workMemory[k].pathList.get(p).get(p2).length]){
						
						// compute position of the instance
						float x=agent.body.getpx();
						float y=agent.body.getpy();
						float rz=agent.body.getrz();
						
						int i=agent.signatureList.workMemory[k].pathList.get(p).get(p2).length-1;
						while (i>=0){
							
							// move forward (>O, >P and >F)
							if (agent.signatureList.workMemory[k].pathList.get(p).get(p2)[i]<=1 || agent.signatureList.workMemory[k].pathList.get(p).get(p2)[i]==3){
								// move forward
								dx= 1*(float)Math.cos(rz);
								dy= 1*(float)Math.sin(rz);
								
								x+=dx;
								y+=dy;
							}
							else if (agent.signatureList.workMemory[k].pathList.get(p).get(p2)[i]==4) rz-=Math.PI/2;
							else if (agent.signatureList.workMemory[k].pathList.get(p).get(p2)[i]==5) rz+=Math.PI/2;

							i--;
						}
						
						// draw instance position
						g.drawOval((int)(40*x+15), (int)(40*(agent.getMain().env.getHeight()-1-y))+15, 10, 10);
			
						dx= 20*(float)Math.cos(rz);
						dy= 20*(float)Math.sin(-rz);
						g.drawLine((int)(40*x+20),    (int)(40*(agent.getMain().env.getHeight()-1-y)+20),
								   (int)(40*x+20+dx), (int)(40*(agent.getMain().env.getHeight()-1-y)+20+dy));
					}
					}
				}
			}
		}
		
		// selectors
		g.setColor(Color.gray);
		g.fillRect(750, 40, 20, 20);
		g.setColor(Color.red);
		g.fillRect(750, 80, 20, 20);
		g.setColor(Color.yellow);
		g.fillRect(750,120, 20, 20);
		g.setColor(Color.green);
		g.fillRect(750,160, 20, 20);
		g.setColor(Color.blue);
		g.fillRect(750,200, 20, 20);
		
		g.setColor(Color.red);
		g.fillRect(800, 80, 20, 20);
		g.setColor(Color.yellow);
		g.fillRect(800,120, 20, 20);
		g.setColor(Color.green);
		g.fillRect(800,160, 20, 20);
		g.setColor(Color.blue);
		g.fillRect(800,200, 20, 20);
		
		g.setColor(Color.black);
		for (short i=0;i<5;i++){
			if (selection[0][i]) g.drawRect(749,39+40*i, 20,20);
			if (selection[1][i]) g.drawRect(799,39+40*i, 20,20);
		}
		
		// distance selector
		for (int i=0;i<Projection.LENGTH+1;i++){
			if (distance[i]) g.setColor(Color.red);
			else g.setColor(Color.black);
			g.fillRect(800, 260+30*i,20,20);
		}
		
		// list type selector
		if (type==0) g.setColor(Color.blue);
		else g.setColor(Color.green);
		g.fillRect(830, 260,20,20);
		
		// hypothesis display selector
		if (hypothesis==0) g.setColor(Color.black);
		else g.setColor(Color.red);
		g.fillRect(830, 300,20,20);
		
		short index=0;
		
		// draw hypothesis
		if (hypothesis==1){
			for (int k=0;k<4;k++){
				if (selection[1][k+1]){
					for (int h=0;h<agent.signatureList.workMemory[k].hypothesisMemory.memory.size();h++){
						
						if (selected_path==index){
							g.setColor(Color.red);
							
							if (change){
								map=agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).getMap();
								change=false;
							}
						}
						else g.setColor(Color.black);
						g.fillRect(885+(int)(index/50)*480, 6+15*(index%50),10,10);
						
						
						g.setColor(Color.black);
						String path="[";
						int i=agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).path.length-1;
						while (i>=0){
							path+=agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).path[i]+", ";
							i--;
						}
						path+="]";
						
						
						g.drawString(path, 900+ (int)(index/50)*480, 15+15*(index%50));
						index++;
						
						for (i=0;i<agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).origin.size();i++){
							for (int n=0;n<agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).origin.get(i).size();n++){
								path="";
								for (int j=agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).origin.get(i).get(n).length-1;j>=0;j--){
									path+=agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).origin.get(i).get(n)[j]+", ";
								}
								
								g.drawString(path, 900+ (int)(index/50)*480, 15+15*(index%50));
								index++;
							}
							
							path="     -> [";
							for (int p=0;p<agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).prototypes.get(i).size();p++){
									path+=agent.signatureList.workMemory[k].hypothesisMemory.memory.get(h).prototypes.get(i).get(p)+", ";
							}
							path+=" ]";
							
							g.drawString(path, 900+ (int)(index/50)*480, 15+15*(index%50));
							index++;
						}
						index++;
					}
					index++;
				}
			}
		}
		
		if (type==0)
		for (int k=0;k<4;k++){
			if (selection[0][k+1]){
				for (int p=0;p<agent.signatureList.workMemory[k].instanceList.size();p++){
					if (distance[agent.signatureList.workMemory[k].instanceList.get(p).length]){

						int i=agent.signatureList.workMemory[k].instanceList.get(p).length-1;
						String path="[";
						while (i>=0){
							path+=agent.signatureList.workMemory[k].instanceList.get(p)[i]+", ";
							i--;
						}
						path+="] - {";
						
						i=0;
						while (i<agent.signatureList.workMemory[k].instanceParts.get(p).size()){
							path+=agent.signatureList.workMemory[k].instanceParts.get(p).get(i)+", ";
							i++;
						}
						path+="}{";
						i=0;
						while (i<agent.signatureList.workMemory[k].instancePattern.get(p).size()){
							path+=agent.signatureList.workMemory[k].instancePattern.get(p).get(i)+", ";
							i++;
						}
						path+="}";
						
						g.setColor(Color.black);
						g.drawString(path, 900+ (int)(index/50)*480, 15+15*(index%50));
						
						if (selected_path==index){
							g.setColor(Color.red);
							
							if (change){
								agent.signatureList.workMemory[k].changeReference(agent.signatureList.workMemory[k].instanceList.get(p), 0);
								change=false;
							}
						}
						else g.setColor(Color.black);
						g.fillRect(885+(int)(index/50)*480, 6+15*(index%50),10,10);
						
						index++;
					}
				}
				index++;
			}
		}
		
		if (type==1){
			for (int k=0;k<4;k++){
				if (selection[0][k+1]){
					for (int p=0;p<agent.signatureList.workMemory[k].movableInstances.size();p++){
						for (int p2=0;p2<agent.signatureList.workMemory[k].movableInstances.get(p).size();p2++){
							
							int i=agent.signatureList.workMemory[k].movableInstances.get(p).get(p2).length-1;
							String path="[";
							while (i>=0){
								path+=agent.signatureList.workMemory[k].movableInstances.get(p).get(p2)[i]+", ";
								i--;
							}
							path+="] - {";
							
							i=0;
							while (i<agent.signatureList.workMemory[k].movableParts.get(p).size()){
								path+=agent.signatureList.workMemory[k].movableParts.get(p).get(i)+", ";
								i++;
							}
							path+="}{";
							i=0;
							while (i<agent.signatureList.workMemory[k].movablePattern.get(p).size()){
								path+=agent.signatureList.workMemory[k].movablePattern.get(p).get(i)+", ";
								i++;
							}
							path+="}";
							g.setColor(Color.black);
							g.drawString(path, 900+ (int)(index/50)*480, 15+15*(index%50));
							
							if (p2==0){
								if (selected_path==index){
									g.setColor(Color.red);
									if (change){

										for (int l=0;l<pattern.length;l++) pattern[l]=0;
										
										//Projection proj=new Projection(agent.signatureList.get(k+InteractionList.nbDF),
										//						 	   agent.signatureList.workMemory[k].movableInstances.get(p).get(0),
										//						 	  agent.signatureList.signatureList,
										//						 	 (byte)-1);
										
										//proj.getResult(pattern);
										
										for (int j=0;j<agent.signatureList.workMemory[k].movableParts.get(p).size();j++) 
											pattern[agent.signatureList.workMemory[k].movableParts.get(p).get(j)]=1;
										
										/*for (int l=0;l<pattern.length;l++){
											if (agent.environmentContext.enactedEnsembles[0][l]==0) pattern[l]=0;
										}*/

										for (int k2=0;k2<4;k2++){
											agent.signatureList.workMemory[k2].changeReference(agent.signatureList.workMemory[k].movableInstances.get(p).get(0), 0);
											agent.signatureList.workMemory[k2].changeReference(agent.signatureList.workMemory[k].movableInstances.get(p).get(1), 1);
										}
										change=false;
									}
								}
								else g.setColor(Color.black);
								g.fillRect(885+(int)(index/50)*480, 6+15*(index%50),10,10);
							}
							
							index++;
						}
						index++;
					}
					index++;
				}
			}
			
			
			// draw pattern
			int offset=0;
			int pixel=5;
			for (int n=0;n<6;n++){
				offset=n*130;
				for (int i=0;i<InteractionList.size;i++){
					for (int j=0;j<InteractionList.size;j++){
						
						g.setColor(new Color(pattern[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+2]*250,
											 pattern[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4  ]*250,
											 pattern[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+3]*250 ));
						
						g.fillRect(1200+i*pixel, (140-j*pixel)+offset, pixel, pixel);
						
						g.setColor(Color.white);
						g.drawOval(1200+12*pixel, (140-12*pixel)+offset, pixel-1, pixel-1);
						
						g.setColor(new Color(pattern[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+1]*250,
											 pattern[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+1]*250,
											 pattern[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+1]*250 ));
			
						g.fillRect(1330+i*pixel, (140-j*pixel)+offset, pixel, pixel);
						
						g.setColor(Color.white);
						g.drawOval(1330+12*pixel, (140-12*pixel)+offset, pixel-1, pixel-1);
					}
				}
			}
			
			// draw map
			offset=0;
			for (int n=0;n<6;n++){
				offset=n*130;
				for (int i=0;i<InteractionList.size;i++){
					for (int j=0;j<InteractionList.size;j++){
						
						g.setColor(new Color(map[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+2]*250,
											 map[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4  ]*250,
											 map[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+3]*250 ));
						
						g.fillRect(1500+i*pixel, (140-j*pixel)+offset, pixel, pixel);
						
						g.setColor(Color.white);
						g.drawOval(1500+12*pixel, (140-12*pixel)+offset, pixel-1, pixel-1);
						
						g.setColor(new Color(map[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+1]*250,
											 map[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+1]*250,
											 map[i*InteractionList.size*4+j*4+n*InteractionList.size*InteractionList.size*4+1]*250 ));
			
						g.fillRect(1630+i*pixel, (140-j*pixel)+offset, pixel, pixel);
						
						g.setColor(Color.white);
						g.drawOval(1630+12*pixel, (140-12*pixel)+offset, pixel-1, pixel-1);
					}
				}
			}
		}
		
		if (type==2){
			for (int k=0;k<4;k++){
				if (selection[0][k+1]){
					for (int p=0;p<agent.signatureList.workMemory[k].movables.size();p++){
							
							int i=agent.signatureList.workMemory[k].movables.get(p).length-1;
							String path="[";
							while (i>=0){
								path+=agent.signatureList.workMemory[k].movables.get(p)[i]+", ";
								i--;
							}
							path+="] ";

							g.setColor(Color.black);
							g.drawString(path, 900+ (int)(index/50)*480, 15+15*(index%50));

						index++;
					}
					index++;
				}
			}
		}
		
		if (type==0)
		for (int k=0;k<4;k++){
			if (selection[1][k+1]){

				for (int p=0;p<agent.signatureList.workMemory[k].pathList.size();p++){
					boolean found=false;
					for (int p2=0;p2<agent.signatureList.workMemory[k].pathList.get(p).size();p2++){

						if (distance[agent.signatureList.workMemory[k].pathList.get(p).get(p2).length]){
							
							int i=agent.signatureList.workMemory[k].pathList.get(p).get(p2).length-1;
							String path="[";
							while (i>=0){
								path+=agent.signatureList.workMemory[k].pathList.get(p).get(p2)[i]+", ";
								i--;
							}
							
							path+="] -> ";
							
							i=agent.signatureList.workMemory[k].originList.get(p).get(p2).length-1;
							while (i>=0){
								path+=agent.signatureList.workMemory[k].originList.get(p).get(p2)[i]+",";
								i--;
							}
							
							g.setColor(Color.black);
							g.drawString(path, 900+ (int)(index/50)*480, 15+15*(index%50));
							
							if (selected_path==index){
								g.setColor(Color.red);
								if (change){
									agent.signatureList.workMemory[k].projectObject(agent.signatureList.workMemory[k].pathList.get(p).get(p2),
												  									agent.signatureList.workMemory[k].originList.get(p).get(p2));
									
									//agent.signatureList.workMemory[k].cleanObjectMemory(agent.environmentContext.enactedEnsembles[0]);
									
									change=false;
								}
							}
							else g.setColor(Color.black);
							g.fillRect(885+(int)(index/50)*480, 6+15*(index%50),10,10);
							
							index++;
							found=true;
						}
					}
					if (found) index++;
				}
				index++;
			}
		}
	}
	
	
	///////////////////////////////////////////////////////////////////////////
	public void mouseClicked(MouseEvent e) {

		clic_x=e.getX();
		clic_y=e.getY();

		// selection of displayed data
		for (short i=0;i<5;i++){
			if (clic_x>=750 && clic_x<=770 && clic_y>=40+40*i && clic_y<=60+40*i) selection[0][i]=!selection[0][i];
		}
		for (short i=0;i<5;i++){
			if (clic_x>=800 && clic_x<=820 && clic_y>=40+40*i && clic_y<=60+40*i) selection[1][i]=!selection[1][i];
		}
		
		for (int i=0;i<Projection.LENGTH+1;i++){
			if (clic_x>=800 && clic_x<=820 && clic_y>=260+30*i && clic_y<=280+30*i) distance[i]=!distance[i];
		}
		
		if (clic_x>=830 && clic_x<=850 && clic_y>=260 && clic_y<=280) type=(byte) ((type+1)%3);
		
		if (clic_x>=830 && clic_x<=850 && clic_y>=300 && clic_y<=320) hypothesis=(byte) ((hypothesis+1)%2);
		
		selected_path=-1;
		change=false;
		
		// selection of path
		if (clic_x>=885){
			int id_x=(clic_x-885)/480;
			int id_y=(clic_y-6)/15;
			if (clic_x>=880+id_x*480 && clic_x<=900+id_x*480 && clic_y>=6+15*id_y && clic_y<=16+15*id_y){
				selected_path=id_x*50+id_y;
				change=true;
			}
		}

		this.repaint();
	}

	@Override
	public void mouseEntered(MouseEvent e){}
	public void mouseExited(MouseEvent e){}
	public void mousePressed(MouseEvent e){}
	public void mouseReleased(MouseEvent e){}
}
