package version10_2_2_1.platform;

import version10_2_2_1.Interface.*;
import version10_2_2_1.display.*;
import version10_2_2_1.environment.*;
import version10_2_2_1.sensorySystem.*;
import version10_2_2_1.agnosticMemory.*;
import version10_2_2_1.Main;

/**
 * whole learning mechanism
 * This class is used as a bridge between modules
 * @author simon gay
 */
public class Agent {

	public Action action;
	public Perception sensors;
	public InteractionList interactionList;
	
	public Robot body;
	public Observer observer;
	public EnvironmentContext environmentContext;
	public SignatureList signatureList;
	
	public NodeList nodeList;
	
	//%%%%%%%%%%%%%%%%% specific to simulators %%%%%%%%%%%%%%%%%%%%%%%
	public TopView topView;
	//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
	public Decision decision;
	
	private Main main;
	
	private int ident;
	
	public Agent(int id, Main m){
		
		action=new Action(this);
		sensors=new Perception(this);
		interactionList=new InteractionList();

		body=new Robot();
		decision=new Decision(this);
		observer=new Observer(this);
		signatureList=new SignatureList(this);
		environmentContext=new EnvironmentContext(this);
		
		nodeList=new NodeList(this);

		//%%%%%%%%%%%%%%%%% specific to simulators %%%%%%%%%%%%%%%%%%%%%%%
		topView=new TopView(this);
		//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

		ident=id;
		
		main=m;
		
	}
	
	// enact a simulation substep
	public void act(){
		action.act();
	}
	
	public boolean isStopped(){
		return action.isStopped();
	}
	
	public int getIdent(){
		return ident;
	}
	
	public Main getMain(){
		return main;
	}
	
	//%%%%%%%%%%%%%%%%% specific to simulators %%%%%%%%%%%%%%%%%%%%%%%
	public void setPosition(float x, float y, float theta){
		body.setPosition(x,y,theta);
		observer.setPosition(x,y);
	}
	
	public void setEnvironment(Environment e){
		body.setEnvironment(e);
		topView.setEnvironment(e);
	}
	//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
	
	// initialize the set of panels related to the agent
	public void setDisplay(){
		
		boolean context=false;
		boolean signatures=true;
		boolean memory=false;
		boolean instances=true;
		boolean pathData=false;
		boolean workMem=true;
		boolean constructible=true;
		
		int size;
		int i;
		boolean found;

		//////////////////////
		if (context){
			size=main.display.nbFrame();
			i=0;
			found=false;
			while (i<size && !found){
				if (main.display.get(i).getClass().getName().equals("version10_2_2_1.display.ContextFrame")) found=true;
				i++;
			}
			if (!found) main.display.addFrame(new ContextFrame(this));
			else
				((ContextFrame)main.display.get(i-1)).setAgent(this);
		}
		
		//////////////////////
		if (signatures){
			size=main.display.nbFrame();
			i=0;
			found=false;
			while (i<size && !found){
				if (main.display.get(i).getClass().getName().equals("version10_2_2_1.display.SignatureFrame")) found=true;
				i++;
			}
			if (!found) main.display.addFrame(new SignatureFrame(this));
			else
				((SignatureFrame)main.display.get(i-1)).setAgent(this);
		}
		
		//////////////////////
		if (memory){
			size=main.display.nbFrame();
			i=0;
			found=false;
			while (i<size && !found){
				if (main.display.get(i).getClass().getName().equals("version10_2_2_1.display.MemoryFrame")) found=true;
				i++;
			}
			if (!found) main.display.addFrame(new MemoryFrame(this));
			else        
				((MemoryFrame)main.display.get(i-1)).setAgent(this);
		}
		
		//////////////////////
		if (instances){
			size=main.display.nbFrame();
			i=0;
			found=false;
			while (i<size && !found){
				if (main.display.get(i).getClass().getName().equals("version10_2_2_1.display.InstanceFrame")) found=true;
				i++;
			}
			if (!found) main.display.addFrame(new InstanceFrame(this));
			else        
				((InstanceFrame)main.display.get(i-1)).setAgent(this);
		}
		
		//////////////////////
		if (pathData){
			size=main.display.nbFrame();
			i=0;
			found=false;
			while (i<size && !found){
				if (main.display.get(i).getClass().getName().equals("version10_2_2_1.display.PathDataFrame")) found=true;
				i++;
			}
			if (!found) main.display.addFrame(new PathDataFrame(this));
			else        
				((PathDataFrame)main.display.get(i-1)).setAgent(this);
		}
		
		//////////////////////
		if (workMem){
			size=main.display.nbFrame();
			i=0;
			found=false;
			while (i<size && !found){
				if (main.display.get(i).getClass().getName().equals("version10_2_2_1.display.WorkMemFrame")) found=true;
				i++;
			}
			if (!found) main.display.addFrame(new WorkMemFrame(this));
			else        
				((WorkMemFrame)main.display.get(i-1)).setAgent(this);
		}
		
		//////////////////////
		if (constructible){
			size=main.display.nbFrame();
			i=0;
			found=false;
			while (i<size && !found){
				if (main.display.get(i).getClass().getName().equals("version10_2_2_1.display.ConstructibleFrame")) found=true;
				i++;
			}
			if (!found) main.display.addFrame(new ConstructibleFrame(this));
			else        
				((ConstructibleFrame)main.display.get(i-1)).setAgent(this);
		}
	}

	// save modules data
	public void save() {
		signatureList.save();
	}
	
	// save images from displayers
	public void saveImage(String path){
		main.saveImage(path);
	}
	
}
